/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1990-1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 */

/*
 * Based on PID S. Ryu, C. Rump and Q. Chunming
 * A predictive and robust active queue management for Internet congestion control 
 * 
 * ISCC 2003. 
 */

#include <math.h>
#include <sys/types.h>
#include "config.h"
#include "template.h"
#include "random.h"
#include "flags.h"
#include "delay.h"
#include "pid3.h"
#include "tcp.h"
#include <iostream>


static class PID3Class : public TclClass {

public:
	PID3Class() : TclClass("Queue/PID3") {}
	TclObject* create(int argc, const char*const* argv) {
		if (argc == 5) 
			return (new PID3Queue(argv[4])) ;
		else
			return (new PID3Queue("Drop")) ;
	}

} class_pid3 ;


PID3Queue::PID3Queue(const char * trace) : CalcTimer(this), link_(NULL), q_(NULL),
	qib_(0), de_drop_(NULL), EDTrace(NULL), tchan_(0), curq_(0),
	aveq_(0), np_(0), newpackets_(0), totalarrivals_(0),
	edp_(), edv_(), first_reset_(1) {

	if (strlen(trace) >= 20) {
		printf(" trace type too long - allocate more space to traceType in pi.h and recompile\n") ;
		exit(0) ;
	}
	strcpy(traceType, trace) ;

	pp = rand() ;

	bind_bool("bytes_", &edp_.bytes) ;	    // boolean: use bytes?
	bind_bool("queue_in_bytes_", &qib_) ;	    // boolean: q in bytes?
	bind("typepid_", &edp_.typepid) ;           // type of discrete pid: 1-"A", 2-"B", 3-"C"
	bind("kp_", &edp_.kp) ; 		
	bind("ki_", &edp_.ki) ;
	bind("kd_", &edp_.kd) ;
	bind("ti_", &edp_.ti) ;
	bind("td_", &edp_.td) ;
	bind_bool("zn_", &edp_.zn) ;
	bind_bool("with_ave_", &edp_.with_ave) ;
	bind("t_", &edp_.t) ;		  
	bind("qref_", &edp_.qref) ;		  
	bind("ewma_weight_", &edp_.ewma_weight) ;
	bind("mean_pktsize_", &edp_.mean_pktsize) ;  // avg pkt size
	bind_bool("setbit_", &edp_.setbit) ;	     // mark instead of drop
	bind("prob_", &edv_.v_prob) ;		     // dropping probability
	bind("curq_", &curq_) ;			     // current queue size
	bind("aveq_", &aveq_) ;		 	     // average queue size
	bind("earlydrops_", &edv_.earlydrops) ;	     // number of dropped packets
	bind("forceddrops_", &edv_.forceddrops) ;    // number of early dropped packets
	bind("newpackets_", &newpackets_) ;
	bind("totalarrivals_", &totalarrivals_) ;
	bind("type_d_b_", &edp_.b) ;
	bind("type_d_n_", &edp_.n) ;

	q_ = new PacketQueue() ;	// underlying queue
	pq_ = q_ ;
	reset() ;

}


void PID3Queue::reset() {

	//double now = Scheduler::instance().clock();
	/*
	if (qib_ && first_reset_ == 1) {
		edp_.qref = edp_.qref*edp_.mean_pktsize;
	}
	*/
	edv_.countini = 0 ;
	edv_.count = 0 ;
	edv_.count_bytes = 0 ;
	edv_.v_prob = 0 ;
	edv_.qold = 0 ;
	edv_.qoldold = 0 ;
	edv_.eold = 0.0 ;
	edv_.eoldold = 0.0 ;
	edv_.i_term_old = 0.0 ;
	edv_.d_term_old = 0.0 ;
	edv_.earlydrops = 0 ;
	edv_.forceddrops = 0 ;
	curq_ = 0 ;
	aveq_ = 0.0 ;
	np_ = 0 ;
	newpackets_ = 0 ;
	totalarrivals_ = 0 ;

	calculate_p() ;
	Queue::reset() ;

}


void PID3Queue::enque(Packet* pkt) {

	//double now = Scheduler::instance().clock();
	hdr_cmn* ch = hdr_cmn::access(pkt) ;
	++edv_.count ;
	edv_.count_bytes += ch->size() ;

	np_++ ;
	totalarrivals_++ ;

	int droptype = DTYPE_NONE ;

	int qlen = qib_ ? q_->byteLength() : q_->length() ;
	curq_ = qlen ;	// helps to trace queue during arrival, if enabled
	aveq_ = (1 - edp_.ewma_weight)*aveq_ + edp_.ewma_weight*qlen ;

	int qlim = qib_ ? (qlim_ * edp_.mean_pktsize) : qlim_ ;

	if (qlen >= qlim) {
		droptype = DTYPE_FORCED ;
	}
	else {
		if (drop_early(pkt, qlen)) {
			droptype = DTYPE_UNFORCED ;
		}
	}

	if (droptype == DTYPE_UNFORCED) {
		edv_.earlydrops++ ;
		Packet *pkt_to_drop = pickPacketForECN(pkt) ;
		if (pkt_to_drop != pkt) {
			q_->enque(pkt) ;
			q_->remove(pkt_to_drop) ;
			pkt = pkt_to_drop ; /* XXX okay because pkt is not needed anymore */
		}

		if (de_drop_ != NULL) {
			if (EDTrace != NULL) 
				((Trace *)EDTrace)->recvOnly(pkt) ;
			de_drop_->recv(pkt) ;
		}
		else {
			drop(pkt) ;
		}
	} 
	else {
		q_->enque(pkt) ;
		if (droptype == DTYPE_FORCED) {
			edv_.forceddrops++ ;
			pkt = pickPacketToDrop() ;
			q_->remove(pkt) ;
			drop(pkt) ;
			edv_.count = 0 ;
			edv_.count_bytes = 0 ;
		}
	}
	return ;

}


double PID3Queue::calculate_p() {

	//double now = Scheduler::instance().clock();
	double p ;
	double qlen ;
	double ki = 0 ;
	double kd = 0 ;
	double e ;
	double p_term = 0, i_term = 0, d_term = 0 ;

	if (edp_.zn) {
		if (edp_.ti) ki = edp_.kp/edp_.ti ;
		kd = edp_.kp * edp_.td ;
	} 
	else {
		ki = edp_.ki ;
		kd = edp_.kd ;
	}

	if (edp_.with_ave) {
		//if (edv_.countini <= 50){
		//	aveq_ = curq_;
		//	edv_.countini++;
		//	std::cout << "Aqui: \n";
		//}
		qlen = qib_ ? aveq_/edp_.mean_pktsize : (double) aveq_ ;
	} 
	else {
		qlen = qib_ ? q_->byteLength()*1.0/edp_.mean_pktsize : q_->length() ;
	}
	e = edp_.qref-qlen ;

	if (edp_.typepid == 1)
		p = edv_.v_prob + edp_.kp*(e-edv_.eold) + ki*edp_.t*e + kd/edp_.t*(e-2*edv_.eold+edv_.eoldold) ;
	else if (edp_.typepid == 2)
		p = edv_.v_prob + edp_.kp*(e-edv_.eold) + ki*edp_.t*e - kd/edp_.t*(qlen-2*edv_.qold+edv_.qoldold) ;
	else if (edp_.typepid == 3)
		p = edv_.v_prob - edp_.kp*(qlen-edv_.qold) + ki*edp_.t*e - kd/edp_.t*(qlen-2*edv_.qold+edv_.qoldold) ;
	else {
		p_term = edp_.kp * (edp_.b*edp_.qref-qlen) ;
		if (edp_.ti) i_term = edv_.i_term_old + edp_.kp * edp_.t * e / edp_.ti ;
		d_term = edp_.td / (edp_.td + edp_.n * edp_.t) * (edv_.d_term_old - edp_.kp * edp_.n * (qlen - edv_.qold)) ;
		p = p_term + i_term + d_term ;
	}

	if (p < 0) p = 0 ;
	if (p > 1) p = 1 ;

	//std::cout << "Ave: " << aveq_ << "; Cur:" << q_->length() << "; p:" << p << "; pp:" << pp << '\n';
	//std::cout << "kp: " << edp_.kp << "; ki:" << ki << "; kd:" << kd << '\n';
	
	edv_.v_prob = p ;
	edv_.qoldold = edv_.qold ;
	edv_.qold = qlen ;
	edv_.eoldold = edv_.eold ;
	edv_.eold = e ;

	edv_.i_term_old = i_term ;
	edv_.d_term_old = d_term ;

	newpackets_ = np_ ;
	np_ = 0 ; 
	CalcTimer.resched(edp_.t) ;
	return p ;

}


int PID3Queue::drop_early(Packet* pkt, int qlen) {

	//double now = Scheduler::instance().clock();
	hdr_cmn* ch = hdr_cmn::access(pkt) ;
	double p = edv_.v_prob ; 

	if (edp_.bytes) {
		p = p*ch->size()/edp_.mean_pktsize ;
		if (p > 1) p = 1 ; 
	}

	double u = Random::uniform() ;
	if (u <= p) {
		edv_.count = 0 ;
		edv_.count_bytes = 0 ;
		hdr_flags* hf = hdr_flags::access(pickPacketForECN(pkt)) ;
		if (edp_.setbit && hf->ect()) {
			hf->ce() = 1 ; 	// mark Congestion Experienced bit
			return (0) ;	// no drop
		} 
		else {
			return (1) ;	// drop
		}
	}
	return (0) ;			// no DROP/mark

}


Packet* PID3Queue::pickPacketForECN(Packet* pkt) {

	return pkt ; /* pick the packet that just arrived */

}


Packet* PID3Queue::pickPacketToDrop() {

	int victim ;
	victim = q_->length() - 1 ;
	return(q_->lookup(victim)) ; 

}


Packet* PID3Queue::deque() {

	Packet *p ;
	p = q_->deque() ;
	curq_ = qib_ ? q_->byteLength() : q_->length() ; // helps to trace queue during arrival, if enabled
	return (p) ;

}


int PID3Queue::command(int argc, const char*const* argv) {

	Tcl& tcl = Tcl::instance() ;
	if (argc == 2) {
		if (strcmp(argv[1], "reset") == 0) {
			reset() ;
			return (TCL_OK) ;
		}
		if (strcmp(argv[1], "early-drop-target") == 0) {
			if (de_drop_ != NULL)
				tcl.resultf("%s", de_drop_->name()) ;
			return (TCL_OK) ;
		}
		if (strcmp(argv[1], "edrop-trace") == 0) {
			if (EDTrace != NULL) {
				tcl.resultf("%s", EDTrace->name()) ;
			}
			else {
				tcl.resultf("0") ;
			}
			return (TCL_OK) ;
		}
		if (strcmp(argv[1], "trace-type") == 0) {
			tcl.resultf("%s", traceType) ;
			return (TCL_OK) ;
		}
	} 
	else if (argc == 3) {
		// attach a file for variable tracing
		if (strcmp(argv[1], "attach") == 0) {
			int mode ;
			const char* id = argv[2] ;
			tchan_ = Tcl_GetChannel(tcl.interp(), (char*)id, &mode) ;
			if (tchan_ == 0) {
				tcl.resultf(" PID3: trace: can't attach %s for writing", id) ;
				return (TCL_ERROR) ;
			}
			return (TCL_OK) ;
		}
		// tell PID3 about link stats
		if (strcmp(argv[1], "link") == 0) {
			LinkDelay* del = (LinkDelay*)TclObject::lookup(argv[2]) ;
			if (del == 0) {
				tcl.resultf(" PID3: no LinkDelay object %s", argv[2]) ;
				return(TCL_ERROR) ;
			}
			link_ = del ;
			return (TCL_OK) ;
		}
		if (strcmp(argv[1], "early-drop-target") == 0) {
			NsObject* p = (NsObject*)TclObject::lookup(argv[2]) ;
			if (p == 0) {
				tcl.resultf("no object %s", argv[2]) ;
				return (TCL_ERROR) ;
			}
			de_drop_ = p ;
			return (TCL_OK) ;
		}
		if (strcmp(argv[1], "edrop-trace") == 0) {
			NsObject * t  = (NsObject *)TclObject::lookup(argv[2]) ;
			if (t == 0) {
				tcl.resultf("no object %s", argv[2]) ;
				return (TCL_ERROR) ;
			}
			EDTrace = t ;
			return (TCL_OK) ;
		}
		if (!strcmp(argv[1], "packetqueue-attach")) {
			delete q_ ;
			if (!(q_ = (PacketQueue*) TclObject::lookup(argv[2])))
				return (TCL_ERROR) ;
			else {
				pq_ = q_ ;
				return (TCL_OK) ;
			}
		}
	}
	return (Queue::command(argc, argv)) ;

}


void PID3Queue::trace(TracedVar* v) {

	char wrk[500], *p ;

	if (((p = strstr(v->name(), "prob")) == NULL) &&
	    ((p = strstr(v->name(), "curq")) == NULL) &&
	    ((p = strstr(v->name(), "aveq")) == NULL) &&
	    ((p = strstr(v->name(), "earlydrops")) == NULL) &&
	    ((p = strstr(v->name(), "forceddrops")) == NULL) &&
	    ((p = strstr(v->name(), "newpackets")) == NULL) &&
	    ((p = strstr(v->name(), "totalarrivals")) == NULL)) {
		fprintf(stderr, " PID3: unknown trace var %s\n", v->name()) ;
		return ;
	}
	
	if (tchan_) {
		int n ;
		double t = Scheduler::instance().clock() ;
		// XXX: be compatible with nsv1 PID3 trace entries
		if (*p == 'c') {
			sprintf(wrk, "%c %g %d", *p, t, int(*((TracedInt*) v))) ;
		} 
		else if ((*p == 'e') || (*p == 'f') || (*p == 'n') || (*p == 't')) {
			sprintf(wrk, "%c %g %d", *p, t, int(*((TracedInt*) v))) ;
		} 
		else {
			sprintf(wrk, "%c %g %g", *p, t, double(*((TracedDouble*) v))) ;
		}
		n = strlen(wrk) ;
		wrk[n] = '\n' ; 
		wrk[n+1] = 0 ;
		(void)Tcl_Write(tchan_, wrk, n+1) ;
	}
	return ; 

}


void PID3CalcTimer::expire(Event *) {

	a_->calculate_p() ;

}
