/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1990-1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


/*
 * Based on PI controller described in:
 * C. Hollot, V. Misra, D. Towsley and W. Gong. 
 * On Designing Improved Controllers for AQM Routers
 * Supporting TCP Flows, 
 * INFOCOMM 2001. 
 */

#ifndef ns_pid3_h
#define ns_pid3_h

#include "queue.h"
#include "trace.h"
#include "timer-handler.h"

#define	DTYPE_NONE	0	/* ok, no drop */
#define	DTYPE_FORCED	1	/* a "forced" drop */
#define	DTYPE_UNFORCED	2	/* an "unforced" (random) drop */


/*
 * Early drop parameters, supplied by user
 */
struct edp_pid3 {
	/*
	 * User supplied.
	 */
	int mean_pktsize ;	/* avg pkt size, linked into Tcl */
	int bytes ;		/* true if queue in bytes, false if packets */
	int setbit ;		/* true to set congestion indication bit */
	int typepid ;		/* type of discrete pid implementation */
	double kp, ki, kd ;	/* parameters to pid controller */
	double ti, td ;		/* Ziegher-Nichols form parameters */
	double b, n ;
	int zn ;		/* bool: parameters in Ziegher-Nichols form? */
	int with_ave ;		/* bool: if set, control average queue instead current */
	double ewma_weight ;
	double t ;		/* sampling frequency (# of times per second) */ 
	double qref ;		/* desired queue size */
	edp_pid3(): mean_pktsize(0), bytes(0), setbit(0), typepid(3), kp(0.0), ki(0.0), kd(0.0), ti(0.0), td(0.0), b(1.0), n(8.0), zn(0), with_ave(0), ewma_weight(0.002), t(1.0), qref(0.0) { }
} ;

/*
 * Early drop variables, maintained by PID3
 */
struct edv_pid3 {
	TracedDouble v_prob ;	/* prob. of packet drop before "count". */
	int countini ;
	int count ;		/* # of packets since last drop */
	int count_bytes ;	/* # of bytes since last drop */
	int qold ;
	int qoldold ;
	double eold ;
	double eoldold ;
	double i_term_old, d_term_old ;
	TracedInt earlydrops ;
	TracedInt forceddrops ;
	edv_pid3() : v_prob(0.0), countini(0), count(0), count_bytes(0), qold(0), qoldold(0), eold(0.0), eoldold(0.0), i_term_old(0.0), d_term_old(0.0), earlydrops(0), forceddrops(0) { }
} ;


class LinkDelay ;
class PID3Queue ; 


class PID3CalcTimer : public TimerHandler {

public:
		PID3CalcTimer(PID3Queue *a) : TimerHandler() { a_ = a; }
		virtual void expire(Event *e) ;

protected:
		PID3Queue *a_ ;

} ;


class PID3Queue : public Queue {
 
 friend class PID3CalcTimer ;

 public:	
	PID3Queue(const char * = "Drop") ;

 protected:
	int pp ;
	int command(int argc, const char*const* argv) ;
	void enque(Packet* pkt) ;
	virtual Packet *pickPacketForECN(Packet* pkt) ;
	virtual Packet *pickPacketToDrop() ;
	Packet* deque() ;
	void reset() ;
	int drop_early(Packet* pkt, int qlen) ;
 	double calculate_p() ;
	PID3CalcTimer CalcTimer ;

	LinkDelay* link_ ;	/* outgoing link */
	int fifo_ ;		/* fifo queue? */
	PacketQueue *q_ ; 	/* underlying (usually) FIFO queue */
		
	int qib_ ;		/* bool: queue measured in bytes? */
	NsObject* de_drop_ ;	/* drop_early target */

	// added to be able to trace EDrop Objects - ratul
	// the other events - forced drop, enque and deque are traced by a different mechanism.
	NsObject * EDTrace ;    // early drop trace
	char traceType[20] ;    // the preferred type for early drop trace, 
	                        // better be less than 19 chars long
	Tcl_Channel tchan_ ;	/* place to write trace records */
	TracedInt curq_ ;	/* current qlen seen by arrivals */
	TracedDouble aveq_ ;	/* average queue size */
	int np_ ;
	TracedInt newpackets_ ;    	/* packet arrivals every sampling time */
	TracedInt totalarrivals_ ; 	/* packet arrivals every sampling time */
	void trace(TracedVar*) ;	/* routine to write trace records */

	edp_pid3 edp_ ;		/* early-drop params */
	edv_pid3 edv_ ;		/* early-drop variables */

	int first_reset_ ;      /* first time reset() is called */

} ;

#endif
